import React, { Component } from 'react';
import { ChromePicker } from 'react-color';
import ReactSlider from 'react-slider';
import PropTypes from 'prop-types';
import Path from './Elements/Path';
import './InteractiveMap.css';
import { TooltipEditor } from './TooltipEditor/TooltipEditor';
import { FilterEditor } from './FilterEditor/FilterEditor';

export class InteractiveMap extends Component {
  static propTypes = {
    imageUrl: PropTypes.string.isRequired,
    handleChange: PropTypes.func.isRequired,
    data: PropTypes.instanceOf(Object).isRequired,
  }

  constructor(props) {
    super(props);
    this.state = {
      selectedNode: null,
      lastPosition: null,
      lastObject: null,
      objects: this.props.data.objects,
      filters: this.props.data.filters,
      selectedColor: '#ff0000',
      selectedFill: 'rgba(0,0,0,0)',
      fillObject: {
        r: 0, g: 0, b: 0, a: 0,
      },
      selectedWidth: 2,
      width: 0,
      height: 0,
    };
  }

  componentDidMount = () => {
    this.getMeta().then((res) => {
      this.setState({ width: res.width, height: res.height });
    });
  }

  getMeta = () => new Promise((resolve) => {
    const img = new Image();
    img.onload = function () {
      resolve({ width: this.width, height: this.height });
    };
    img.src = this.props.imageUrl;
  })

  getTags = () => {
    if (this.state.lastObject === null) {
      return <span className="tags-select">Please select an object to add tags.</span>;
    }
    const { tags } = this.state.objects[this.state.lastObject];
    const tagItems = tags.map(tag => (
      <div className="tag">
        <span className="name">{`${tag.question} : ${tag.answer}`}</span>
        <button onClick={() => { this.handleToggleTag(tag); }} className="close">x</button>
      </div>
    ));
    tagItems.push((
      <div className="add-tag">
        <input type="text" value={this.state.addTagQuestionValue} onChange={this.handleAddTagQuestionChange} />
        <input type="text" value={this.state.addTagAnswerValue} onChange={this.handleAddTagAnswerChange} />
        <button onClick={() => { this.handleToggleTag(null); }}>Add</button>
      </div>
    ));
    return tagItems;
  }

  getIcons = () => {
    if (this.state.lastObject === null) {
      return <span className="icons-select">Please select an object to add icons.</span>;
    }
    return ((
      <div className="add-icon">
        <input placeholder="Enter the URL" type="text" value={this.state.iconUrl} onChange={this.handleIconUrlChange} />
        <button onClick={this.handleAddIcon}>Add</button>
      </div>));
  }

  clickCanAddNewPoint = (target) => {
    if ((target.className.baseVal === 'path-line' && this.state.lastObject === null) || target.className.baseVal === 'svg-icon' || target.className.baseVal === 'bezier-handler' || target.className.baseVal === 'start-handler') {
      /* We don't create a new point if we:
        - click on the drag handle
        - click on the finish point (yellow)
        - click on an icon
        - click on the path when there is no selected object */
      return false;
    }
    return true;
  }

  handleIconUrlChange = (event) => {
    this.setState({ iconUrl: event.target.value });
  }

  handleToggleTag = (tag) => {
    const tagToToggle = tag || { question: this.state.addTagQuestionValue, answer: this.state.addTagAnswerValue };
    const { objects } = this.state;
    const { tags } = objects[this.state.lastObject];
    if (tags.includes(tagToToggle)) {
      tags.splice(tags.indexOf(tagToToggle), 1);
    } else {
      tags.push(tagToToggle);
    }
    objects[this.state.lastObject].tags = tags;
    this.setState({
      objects,
      addTagQuestionValue: tag ? this.state.addTagQuestionValue : '',
      addTagAnswerValue: tag ? this.state.addTagAnswerValue : '',
    });
  }

  handleAddIcon = () => {
    const { objects } = this.state;
    const { icons } = objects[this.state.lastObject];
    icons.push({
      url: this.state.iconUrl,
      x: 50,
      y: 50,
    });
    objects[this.state.lastObject].icons = icons;
    this.setState({ objects, iconUrl: '' });
  }

  handleAddTagQuestionChange = (event) => {
    this.setState({ addTagQuestionValue: event.target.value });
  }

  handleAddTagAnswerChange = (event) => {
    this.setState({ addTagAnswerValue: event.target.value });
  }

  handleMouseMove = ({ clientX, clientY }) => {
    const { selectedNode } = this.state;

    if (selectedNode === null) {
      return;
    }
    const { objects } = this.state;
    if (selectedNode.data > -1) { // Moving control points
      const svgRect = this.map.getBoundingClientRect();
      const svgX = clientX - svgRect.left;
      const svgY = clientY - svgRect.top;
      const { data } = objects[selectedNode.object];
      data[selectedNode.data].controlX = svgX;
      data[selectedNode.data].controlY = svgY;
      objects[selectedNode.object].data = data;
    } else if (selectedNode.icon > -1) { // Moving icons
      const svgRect = this.map.getBoundingClientRect();
      const svgX = clientX - svgRect.left;
      const svgY = clientY - svgRect.top;
      const { icons } = objects[selectedNode.object];
      icons[selectedNode.icon].x = svgX;
      icons[selectedNode.icon].y = svgY;
      objects[selectedNode.object].icons = icons;
    }
    this.setState({
      objects,
    });
  }

  handleMouseDown = (selectedNode) => {
    this.setState({ selectedNode });
  }

  handleMouseUp = () => {
    this.setState({ selectedNode: null });
  }

  handleNewPoint = ({ clientX, clientY, target }) => {
    if (!this.clickCanAddNewPoint(target)) {
      return;
    }
    let { lastPosition, lastObject } = this.state;
    const svgRect = this.map.getBoundingClientRect();
    const svgX = clientX - svgRect.left;
    const svgY = clientY - svgRect.top;
    const { objects } = this.state;
    if (lastPosition === null) {
      lastPosition = { x: svgX, y: svgY };
      this.setState({ lastPosition });
      return;
    }
    const newData = {
      startX: lastPosition.x,
      startY: lastPosition.y,
      controlX: (svgX + lastPosition.x) / 2,
      controlY: (svgY + lastPosition.y) / 2,
      endX: svgX,
      endY: svgY,
    };
    lastPosition = { x: svgX, y: svgY };
    if (lastObject === null) {
      objects.push({
        tags: [],
        icons: [],
        tooltip: {},
        color: this.state.selectedColor,
        fill: this.state.selectedFill,
        width: this.state.selectedWidth,
        data: [newData],
      });
      lastObject = objects.length - 1;
    } else {
      const { data } = objects[lastObject];
      data.push(newData);
      objects[lastObject].data = data;
    }
    this.setState({ objects, lastObject, lastPosition });
  }

  handleClose = (svgX, svgY) => {
    const { lastPosition, lastObject, objects } = this.state;
    const newData = {
      startX: lastPosition.x,
      startY: lastPosition.y,
      controlX: (svgX + lastPosition.x) / 2,
      controlY: (svgY + lastPosition.y) / 2,
      endX: svgX,
      endY: svgY,
    };
    const { data } = objects[lastObject];
    data.push(newData);
    objects[lastObject].data = data;
    objects[lastObject].finished = true;
    this.setState({ objects, lastPosition: null, lastObject: null });
  }

  handleStop = (event) => {
    if (event.key !== 'f') {
      return;
    }
    this.setState({ lastObject: null, lastPosition: null });
  }

  handleSelectObject = (objectIndex) => {
    this.setState({ lastObject: objectIndex });
  }

  handleRemoveLastIcon = () => {
    if (!(this.state.lastObject !== null)) {
      return;
    }
    const { objects } = this.state;
    const { icons } = objects[this.state.lastObject];
    if (icons.length === 0) {
      return;
    }
    icons.splice(icons.length - 1, 1);
    objects[this.state.lastObject].icons = icons;
    this.setState({ objects });
  }

  handleRemoveLastLine = () => {
    if (!(this.state.lastPosition && this.state.lastObject !== null)) {
      return;
    }
    const { objects } = this.state;
    const { data } = objects[this.state.lastObject];
    data.splice(data.length - 1, 1);
    objects[this.state.lastObject].data = data;
    if (data.length === 0) {
      this.setState({ objects }, () => {
        this.handleRemoveSelectedObject();
      });
    } else {
      const newLastPosition = { x: data[data.length - 1].endX, y: data[data.length - 1].endY };
      this.setState({ objects, lastPosition: newLastPosition });
    }
  }

  handleRemoveSelectedObject = () => {
    if (this.state.lastObject === null) {
      return;
    }
    const { objects } = this.state;
    objects.splice(this.state.lastObject, 1);
    this.setState({
      lastObject: null,
      lastPosition: null,
      objects,
      tooltipEditOpen: false,
      filterEditOpen: false,
    });
  }

  handleFillChange = (value) => {
    this.setState({ fillObject: value.rgb, selectedFill: `rgba(${value.rgb.r},${value.rgb.b},${value.rgb.g},${value.rgb.a})` });
  }

  handleFillPickerOpen = () => {
    this.setState({ openFillPicker: !this.state.openFillPicker });
  }

  handleColorChange = ({ hex }) => {
    this.setState({ selectedColor: hex });
  }

  handleColorPickerOpen = () => {
    this.setState({ openColorPicker: !this.state.openColorPicker });
  }

  handleWidthPickerOpen = () => {
    this.setState({ openWidthPicker: !this.state.openWidthPicker });
  }

  handleWidthChange = (selectedWidth) => {
    this.setState({ selectedWidth });
  }

  handleTagsOpen = () => {
    this.setState({ openTags: !this.state.openTags });
  }

  handleIconsOpen = () => {
    this.setState({ openIcons: !this.state.openIcons });
  }

  handleTooltipEditOpen = () => {
    if (this.state.lastObject !== null) {
      this.setState({ filterEditOpen: false, tooltipEditOpen: !this.state.tooltipEditOpen });
    }
  }

  handleFilterEditOpen = () => {
    this.setState({ tooltipEditOpen: false, filterEditOpen: !this.state.filterEditOpen });
  }

  handleObjectChange = (object, index) => {
    const { objects } = this.state;
    objects[index] = object;
    this.setState({ objects });
  }

  handleFiltersChange = (filters) => {
    this.setState({ filters });
  }

  handleSave = () => {
    const { objects } = this.state;
    for (let i = 0; i < objects.length; i += 1) {
      const filledPath = this.generateFilledPath(objects[i].data);
      objects[i].filledPath = filledPath;
    }
    const data = {
      objects,
      filters: this.state.filters,
      viewbox: `0 0 ${this.state.width} ${this.state.height}`,
    };
    this.props.handleChange(data);
  }

  generateFilledPath = (data) => {
    let dataString = '';

    for (let i = 0; i < data.length; i += 1) {
      if (i === 0) {
        dataString += `M${data[i].startX},${data[i].startY} Q ${data[i].controlX},${data[i].controlY} ${data[i].endX},${data[i].endY}`;
      } else if (data[i].startX === data[i - 1].endX && data[i].startY === data[i - 1].endY) {
        dataString += ` Q ${data[i].controlX},${data[i].controlY} ${data[i].endX},${data[i].endY}`;
      } else {
        dataString += `M${data[i].startX},${data[i].startY} Q ${data[i].controlX},${data[i].controlY} ${data[i].endX},${data[i].endY}`;
      }
    }
    return dataString;
  }

  render() {
    return (
      <div className="map-wrapper">
        <div className="map-svg">
          {this.state.tooltipEditOpen &&
            <TooltipEditor
              object={this.state.objects[this.state.lastObject]}
              onChange={(object) => { this.handleObjectChange(object, this.state.lastObject); }}
              width={this.state.width}
              height={this.state.height}
            />
          }

          {this.state.filterEditOpen &&
          <FilterEditor
            filters={this.state.filters}
            onChange={this.handleFiltersChange}
            width={this.state.width}
            height={this.state.height}
          />
          }

          {!this.state.filterEditOpen && !this.state.tooltipEditOpen &&
          <svg
            ref={(node) => { this.map = node; }}
            tabIndex="0"
            viewbox={`0 0 ${this.state.width} ${this.state.height}`}
            height={this.state.height}
            width={this.state.width}
            onKeyDown={event => this.handleStop(event)}
            onMouseMove={event => this.handleMouseMove(event)}
            onMouseUp={() => this.handleMouseUp()}
            onMouseDown={event => this.handleNewPoint(event)}
          >
            <image xlinkHref={this.props.imageUrl} height="100%" width="100%" />
            {this.state.objects.map((object, objectIndex) => {
            const paths = object.data.map((path, index) => (
              <Path
                start={index === 0 && this.state.lastObject === objectIndex && !object.finished}
                selected={this.state.lastObject === objectIndex}
                startX={path.startX}
                startY={path.startY}
                controlX={path.controlX}
                controlY={path.controlY}
                endX={path.endX}
                endY={path.endY}
                color={object.color}
                width={object.width}
                handleSelectObject={() => { this.handleSelectObject(objectIndex); }}
                handleClose={() => { this.handleClose(path.startX, path.startY); }}
                handleMouseDown={() => { this.handleMouseDown({ object: objectIndex, data: index }); }}
              />
            ));
            const icons = object.icons.map((icon, index) => (
              <image onClick={() => { this.handleSelectObject(objectIndex); }} onMouseDown={() => { this.handleMouseDown({ object: objectIndex, icon: index }); }} className="svg-icon" xlinkHref={icon.url} transform="translate(-16,-16)" width="32" height="32" x={icon.x} y={icon.y} />
            ));
            return [paths, icons];
            })}
            {this.state.lastPosition &&
            <circle cx={this.state.lastPosition.x} cy={this.state.lastPosition.y} r="5" stroke="black" fill="cyan" />
          }
          </svg>
          }


        </div>

        <div className="map-controls">
          <button onClick={this.handleRemoveLastLine}>Remove last line</button>
          <button onClick={this.handleRemoveLastIcon}>Remove last icon</button>
          <button onClick={this.handleRemoveSelectedObject}>Remove selected object</button>
          <button onClick={this.handleFillPickerOpen}>Fill adjust</button>
          <div className={`color-picker ${this.state.openFillPicker ? 'open' : ''}`}>
            <ChromePicker
              color={this.state.fillObject}
              onChangeComplete={this.handleFillChange}
            />
          </div>
          <button onClick={this.handleColorPickerOpen}>Color adjust</button>
          <div className={`color-picker ${this.state.openColorPicker ? 'open' : ''}`}>
            <ChromePicker
              color={this.state.selectedColor}
              onChangeComplete={this.handleColorChange}
            />
          </div>
          <button onClick={this.handleWidthPickerOpen}>Width adjust</button>
          <div className={`width-picker ${this.state.openWidthPicker ? 'open' : ''}`}>
            <span>{this.state.selectedWidth}</span>
            <ReactSlider
              value={this.state.selectedWidth}
              defaultValue={this.state.selectedWidth}
              min={1}
              max={20}
              onChange={this.handleWidthChange}
            />

          </div>
          <button onClick={this.handleTagsOpen}>Add tags</button>
          <div className={`tags-wrapper ${this.state.openTags ? 'open' : ''}`}>
            {this.getTags()}
          </div>
          <button onClick={this.handleIconsOpen}>Add icons</button>
          <div className={`icons-wrapper ${this.state.openIcons ? 'open' : ''}`}>
            {this.getIcons()}
          </div>
          <button onClick={this.handleTooltipEditOpen}>Edit tooltip</button>
          <button onClick={this.handleFilterEditOpen}>Edit filters</button>
          <button onClick={() => { this.handleStop({ key: 'f' }); }}>Finish (F key) </button>
          <button onClick={this.handleSave}>Save changes</button>

        </div>
      </div>

    );
  }
}

export default InteractiveMap;
